var stream = require('stream');
var crypto = require('crypto');
var fs = require('fs');

var express = require('express');
var bodyParser = require('body-parser');
var nodemailer = require('nodemailer');
var directTransport = require('nodemailer-direct-transport');
var mime = require('mime');
var parseHeader = require('parse-http-header');
var UriTemplate = require('uri-templates');

// Pipe through to another stream without changes
// - otherwise the "path" property screws things up when using the request object/stream
function copyStream(source) {
	var content = new stream.Transform();
	content._transform = function (chunk, encoding, callback) {
		this.push(chunk, encoding);
		callback();
	};
	source.pipe(content);
	return content;
}

function commaSplit(string) {
	return string.match(/([^,]|"([^"]|\\.)*")*/g);
}

function htmlEscape(value) {
	return value.replace(/&/g, '&amp;').replace(/</g, '&lt;');
}

function Transform(obj) {
	var emailTemplate = UriTemplate(obj.email);
	var urlTemplates = [].concat(obj.urls || obj.url || []).map(UriTemplate);
	var name = obj.name;
	var open = (obj.open || obj.open == undefined);

	this.emailForUrl = function (url, mustBeOpen) {
		if (mustBeOpen && !open) return null;
		for (var i = 0; i < urlTemplates.length; i++) {
			var params = urlTemplates[i].fromUri(url);
			if (params) return emailTemplate.fill(params);
		}
		return null;
	};
	this.urlsForEmail = function (email, mustBeOpen) {
		if (mustBeOpen && !open) return [];
		var params = emailTemplate.fromUri(email);
		if (params) return urlTemplates.map(function (template) {
			return template.fill(params);
		});
		return [];
	};
	this.nameForUrl = function (url, mustBeOpen) {
		var email = this.emailForUrl(url, mustBeOpen);
		if (!email) return null;
		if (typeof name === 'function') {
			return name(url, email);
		} else {
			return name || url;
		}
	};
}
Transform.prototype = {};

function Mapping(transforms) {
	var transforms = [].concat(transforms).map(function (transform) {
		return new Transform(transform);
	});
	this.emailForUrl = function (url, mustBeOpen) {
		for (var i = 0; i < transforms.length; i++) {
			var email = transforms[i].emailForUrl(url, mustBeOpen);
			if (email) return email;
		}
		return null;
	};
	this.urlsForEmail = function (email, mustBeOpen) {
		var urls = [];
		for (var i = 0; i < transforms.length; i++) {
			urls = urls.concat(transforms[i].urlsForEmail(email, mustBeOpen));
		}
		return urls;
	};
	this.nameForUrl = function (url, mustBeOpen) {
		for (var i = 0; i < transforms.length; i++) {
			var name = transforms[i].nameForUrl(url, mustBeOpen);
			if (name) return name;
		}
		return null;
	};
}

function httpMail(options) {
	options = options || {};
	var maxInline = options.maxInline || 500*1024; // 500kb maximum inline document
	var mapping = new Mapping(options.mapping);
	var mailTransport = nodemailer.createTransport(options.transport || directTransport());
	mailTransport.on('log', function (message) {
		console.log(message.type + ': ' + message.message);
	});

	// HTML generator for GET requests (web form)
	var htmlFunction;
	if (options.webForm !== false) {
		if (!options.webForm || options.webForm === true) {
			options.webForm = __dirname + '/form-template.html';
		}
		if (typeof options.webForm === 'string') {
			var template;
			try {
				template = fs.readFileSync(options.webForm, {encoding: 'utf-8'});
			} catch (e) {
				template = '<B>Error</B>: could not load template';
			}
			htmlFunction = function (username, url) {
				return template.replace(/\{\{url\}\}/g, htmlEscape(url)).replace(/\{\{name\}\}/g, htmlEscape(username));
			};
		} else if (typeof options.webForm === 'function') {
			htmlFunction = options.webForm;
		} else {
			htmlFunction = function () {
				return '<B>Error</B>: invalid template argument: ' + JSON.stringify(options.webForm).replace(/&/g, '&amp;').replace(/</g, '&lt;');
			};
		}
	};

	var router = express.Router({strict: true});
	router.use('',
		function (request, response, next) {
			// Detect email address, or reject
			var url = request.originalUrl;
			var mappedEmail = request.mappedEmail = mapping.emailForUrl(url);
			request.mappedName = mapping.nameForUrl(url) || url;
			if (!mappedEmail) return next('route'); // Skip CORS and bodyParser stuff
			next();
		},
		function (request, response, next) {
			response.header("Access-Control-Allow-Origin", "*");
			//response.header("Access-Control-Request-Method", "GET, POST, OPTIONS");
			response.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept, Link, From, Subject");
			next();
		},
		bodyParser.urlencoded({extended: false}),
		function (request, response, next) {
			var email = request.mappedEmail, displayName = request.mappedName;
			if (request.method === 'GET' && htmlFunction) {
				response.setHeader('Content-Type', 'text/html');
				var email = request.mappedEmail, displayName = request.mappedName;
				if (!displayName || !email) return next();

				var url = request.protocol + '://' + request.get('host') + request.originalUrl;
				response.send(htmlFunction(displayName, url));
			} else if (request.method === 'POST') {
				if (!request.headers['from']) return response.status(400).json('"From" is required');

				var contentType = request.headers['content-type'];
				var filename = 'attachment', cid;
				var inline = !(parseFloat(request.headers['content-length']) > options.maxInline);
				if (request.headers['content-disposition']) {
					var parsed = parseHeader(request.headers['content-disposition']);
					filename = parsed.filename || filename;
					inline = inline && (parsed[0] == 'inline');
				}

				// Envelope sender/target
				var sender = (request.headers['sender'] || commaSplit(request.headers['from'])[0]);
				var target = email;
				// strip name and <>
				sender = sender.replace(/^\s*"([^"\\]|\\.)*"/, '').replace(/(^[^<]*<|>[^>]*$)/g, '');
				target = target.replace(/^\s*"([^"\\]|\\.)*"/, '').replace(/(^[^<]*<|>[^>]*$)/g, '');

				var mail = {
					from: request.headers['from'],
					to: [email],
					cc: [],
					subject: request.headers['subject'] || '',
					text: '',
					headers: {},
					envelope: { /* only send to the "to" address, don't include CC/BCC */
						from: sender,
						to: target
					}
				};
				if (request.headers['mail-cc']) {
					mail.cc = mail.cc.concat(commaSplit(request.headers['mail-cc']));
				}
				if (request.headers['cc']) {
					mail.cc = mail.cc.concat(commaSplit(request.headers['cc']));
				}

				// TODO: copy over all suitable headers from request

				if (inline && /^text\/plain($|\s|;)/.test(contentType)) {
					mail.text = copyStream(request);
				} else if (inline && /^text\/html($|\s|;)/.test(contentType)) {
					delete mail.text;
					mail.html = copyStream(request);
				} else {
					if (/^image\//.test(contentType) && inline) {
						delete mail.text;
						cid = crypto.pseudoRandomBytes(16).toString('hex') + '-' + email;
						mail.html = '<img src="cid:' + cid + '" style="max-width: 100%; max-height: 100%" alt="' + filename.replace(/&/g, '&amp;').replace(/"/g, '&quot;').replace(/</g, '&lt') + '"/>';
					}
					mail.attachments = [{
						contentType: contentType,
						content: copyStream(request),
						cid: cid,
						filename: filename
					}];
				}

				// Replace all URLs in headers
				function transformField(value) {
					return value.replace(/<([^<>]+\:\/\/[^<>]+)>/g, function (match, url) {
						var email = mapping.emailForUrl(url);
						if (email) return '<' + email + '>';
						return match;
					});
				}
				for (var key in mail.headers) {
					mail.headers[key] = transformField(mail.headers[key]);
				}
				mail.from = transformField(mail.from);
				mail.cc = mail.cc.map(transformField);

				mailTransport.sendMail(mail, function (error, info) {
					if (error) {
						return response.status(500).json({
							code: error.code,
							error: error.message
						});
					}
					return response.json('Message accepted');
				});
			} else {
				next();
			}
		}
	);
	return router;
}

module.exports = httpMail;
